#!/usr/bin/env python
"""Test `omega.fol_bdd`."""
from omega.symbolic import symbolic


def test_quantifiers():
    """Test quantifier syntax in LTL."""
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='int', dom=(0, 5), owner='sys'))
    aut = a.build()
    # TYPENESS !!!
    s = '\E x: (x = 7)'
    u = aut.add_expr(s)
    print(u)


if __name__ == '__main__':
    test_quantifiers()



"""Algorithms for generalized Streett and Rabin games.


References
==========

Roderick Bloem, Barbara Jobstmann, Nir Piterman,
Amir Pnueli, Yaniv Sa'ar
    "Synthesis of reactive(1) designs"
    Journal of Computer and System Sciences
    Vol.78, No.3, pp.911--938, 2012


Robert Konighofer
    "Debugging formal specifications with
    simplified counterstrategies"
    Master's thesis
    Inst. for Applied Information Processing and Communications,
    Graz University of Technology, 2009
"""
import logging
import copy
from omega.symbolic import fixpoint as fx
from omega.symbolic import symbolic


logger = logging.getLogger(__name__)


# TODO:
# - switch to `dd.autoref`


def solve_streett_game(aut, rank=1):
    """Return winning set and iterants for Streett(1) game.

    @param aut: compiled game with <>[] | []<> winning
    @type aut: `symbolic.Automaton`
    """
    assert rank == 1, 'only rank 1 supported for now'
    assert not aut.vars or aut.bdd.vars, (
        'first call `Automaton.build`')
    aut.assert_consistent(built=True)
    assert len(aut.win['<>[]']) > 0
    assert len(aut.win['[]<>']) > 0
    bdd = aut.bdd
    z = bdd.true
    zold = None
    while z != zold:
        zold = z
        xijk = list()
        yij = list()
        for goal in aut.win['[]<>']:
            y, yj, xjk = _attractor_under_assumptions(z, goal, aut)
            z = bdd.apply('and', z, y)
            xijk.append(xjk)
            yij.append(yj)
    return z, yij, xijk


def _attractor_under_assumptions(z, goal, aut):
    """Targeting `goal`, under unconditional assumptions."""
    bdd = aut.bdd
    env_action = aut.action['env'][0]
    sys_action = aut.action['sys'][0]
    xjk = list()
    yj = list()
    y = bdd.false
    yold = None
    cox_z = fx.ue_preimage(env_action, sys_action, z, aut)
    g = bdd.apply('and', goal, cox_z)
    while y != yold:
        yold = y
        cox_y = fx.ue_preimage(env_action, sys_action, y, aut)
        unless = bdd.apply('or', cox_y, g)
        xk = list()
        for safe in aut.win['<>[]']:
            x = fx.trap(env_action, sys_action,
                        safe, aut, unless=unless)
            xk.append(x)
            y = bdd.apply('or', y, x)
        yj.append(y)
        xjk.append(xk)
    return y, yj, xjk


def make_streett_transducer(z, yij, xijk, aut):
    """Return I/O `symbolic.Automaton` implementing strategy.

    An auxiliary variable `_goal` is added,
    to represent the counter of recurrence goals.
    """
    aut.assert_consistent(built=True)
    assert z != aut.bdd.false, 'empty winning set'
    # add goal counter var
    c = '_goal'
    dvars = copy.deepcopy(aut.vars)
    n_goals = len(aut.win['[]<>'])
    dvars[c] = dict(
        type='saturating',
        dom=(0, n_goals - 1),
        owner='sys')
    # compile transducer with refined shared BDD
    t = symbolic.Automaton()
    t.vars = dvars
    bdd = aut.bdd
    t.bdd = bdd
    t = t.build()
    env_init = aut.init['env'][0]
    sys_init = aut.init['sys'][0]
    env_action = aut.action['env'][0]
    sys_action = aut.action['sys'][0]
    holds = aut.win['<>[]']
    goals = aut.win['[]<>']
    # compute strategy from iterates
    # \rho_1: switch goals
    rho_1 = bdd.false
    for i, goal in enumerate(goals):
        ip = (i + 1) % len(goals)
        s = "({c} = {i}) & ({c}' = {ip})".format(c=c, i=i, ip=ip)
        u = t.add_expr(s)
        u = bdd.apply('and', u, goal)
        rho_1 = bdd.apply('or', u, rho_1)
    zp = bdd.rename(z, t.prime)
    rho_1 = bdd.apply('and', rho_1, zp)
    # \rho_2: descent in basin
    rho_2 = bdd.false
    for i, yj in enumerate(yij):
        s = "({c} = {i}) & ({c}' = {i})".format(c=c, i=i)
        count = t.add_expr(s)
        rho_2j = bdd.false
        basin = yj[0]
        for y in yj[1:]:
            next_basin = bdd.rename(basin, t.prime)
            rim = bdd.apply('diff', y, basin)
            u = bdd.apply('and', rim, next_basin)
            rho_2j = bdd.apply('or', rho_2j, u)
            basin = bdd.apply('or', basin, y)
        u = bdd.apply('and', rho_2j, count)
        rho_2 = bdd.apply('or', rho_2, u)
    # \rho_3: persistence holds
    rho_3 = bdd.false
    for i, xjk in enumerate(xijk):
        s = "({c} = {i}) & ({c}' = {i})".format(c=c, i=i)
        count = t.add_expr(s)
        rho_3j = bdd.false
        used = bdd.false
        for xk in xjk:
            assert len(xk) == len(holds), xk
            for x, hold in zip(xk, holds):
                next_wait = bdd.rename(x, t.prime)
                stay = bdd.apply('diff', x, used)
                used = bdd.apply('or', used, x)
                u = bdd.apply('and', stay, next_wait)
                u = bdd.apply('and', u, hold)
                rho_3j = bdd.apply('or', rho_3j, u)
        u = bdd.apply('and', rho_3j, count)
        rho_3 = bdd.apply('or', rho_3, u)
    # \rho
    u = bdd.apply('or', rho_1, rho_2)
    u = bdd.apply('or', rho_3, u)
    u = bdd.apply('and', sys_action, u)
    # counter `c` limits
    u = bdd.apply('and', t.action['sys'][0], u)
    if not aut.plus_one:
        u = bdd.apply('->', env_action, u)
    if aut.moore:
        u = bdd.forall(t.upvars, u)
    t.action['sys'] = [u]
    # initial condition for counter
    # (no closure taken for counter)
    s = '{c} = 0'.format(c=c)
    count = t.add_expr(s)
    win_set = z
    # \A init
    init = _all_init(env_init, sys_init, count, bdd)
    t.init['env'] = [init]
    init = bdd.apply('->', init, win_set)
    assert init == bdd.true, 'losing from some init states'
    return t


def solve_rabin_game(aut, rank=1):
    """Return winning set and iterants for Rabin(1) game.

    @param aut: compiled game with <>[] & []<> winning
    @type aut: `symbolic.Automaton`
    """
    assert rank == 1, 'only rank 1 supported for now'
    assert not aut.vars or aut.bdd.vars, (
        'first call `Automaton.build`')
    aut.assert_consistent(built=True)
    # TODO: can these assertions be removed elegantly ?
    assert len(aut.win['<>[]']) > 0
    assert len(aut.win['[]<>']) > 0
    bdd = aut.bdd
    z = bdd.false
    zold = None
    zk = list()
    yki = list()
    xkijr = list()
    while z != zold:
        zold = z
        xijr = list()
        yi = list()
        for hold in aut.win['<>[]']:
            y, xjr = _cycle_inside(zold, hold, aut)
            z = bdd.apply('or', z, y)
            xijr.append(xjr)
            yi.append(y)
        zk.append(z)
        yki.append(yi)
        xkijr.append(xijr)
    return zk, yki, xkijr


def _cycle_inside(z, hold, aut):
    """Cycling through goals, while staying in `hold`."""
    bdd = aut.bdd
    env_action = aut.action['env'][0]
    sys_action = aut.action['sys'][0]
    cox_z = fx.ue_preimage(env_action, sys_action,
                           z, aut)
    g = bdd.apply('or', cox_z, hold)
    y = bdd.true
    yold = None
    while y != yold:
        yold = y
        cox_y = fx.ue_preimage(env_action, sys_action,
                               y, aut)
        inside = bdd.apply('and', cox_y, g)
        xjr = list()
        for goal in aut.win['[]<>']:
            x, xr = _attractor_inside(inside, goal, aut)
            xjr.append(xr)
            y = bdd.apply('and', y, x)
    return y, xjr


def _attractor_inside(inside, goal, aut):
    bdd = aut.bdd
    env_action = aut.action['env'][0]
    sys_action = aut.action['sys'][0]
    xr = list()
    x = bdd.false
    xold = None
    while x != xold:
        xold = x
        cox_x = fx.ue_preimage(
            env_action, sys_action, x, aut,
            evars=aut.epvars)
        x = bdd.apply('or', cox_x, goal)
        x = bdd.apply('and', x, inside)
        x = bdd.apply('or', x, xold)
        xr.append(x)
    return x, xr


def make_rabin_transducer(zk, yki, xkijr, aut):
    """Return O/I transducer for Rabin(1) game."""
    aut.assert_consistent(built=True)
    win_set = zk[-1]
    assert win_set != aut.bdd.false, 'empty winning set'
    dvars = dict(aut.vars)
    n_holds = len(aut.win['<>[]'])
    n_goals = len(aut.win['[]<>'])
    # add transducer memory as two indices:
    # - `w`: persistence hold index
    # - `c`: recurrence goal index
    w = '_hold'
    c = '_goal'
    n_w = n_holds - 1 + 1  # last value used as "none"
    n_c = n_goals - 1
    dvars[w] = dict(type='saturating', dom=(0, n_w), owner='sys')
    dvars[c] = dict(type='saturating', dom=(0, n_c), owner='sys')
    # compile
    t = symbolic.Automaton()
    t.vars = dvars
    bdd = aut.bdd
    t.bdd = bdd
    t = t.build()
    env_init = aut.init['env'][0]
    sys_init = aut.init['sys'][0]
    env_action = aut.action['env'][0]
    sys_action = aut.action['sys'][0]
    goals = aut.win['[]<>']
    t.action['env'] = [env_action]
    # compute strategy from iterates
    # \rho_1: descent in persistence basin
    s = "({c}' = {c}) & ({w}' = {none})".format(
        c=c, w=w, none=n_holds)
    count = t.add_expr(s)
    rho_1 = bdd.false
    basin = zk[0]
    for z in zk[1:]:
        trans = _moore_trans(basin, t)
        rim = bdd.apply('diff', z, basin)
        u = bdd.apply('and', rim, trans)
        u = bdd.apply('and', u, count)
        rho_1 = bdd.apply('or', rho_1, u)
        basin = z
    rho_2 = bdd.false
    rho_3 = bdd.false
    rho_4 = bdd.false
    basin = bdd.false
    for z, yi, xijr in zip(zk, yki, xkijr):
        cox_basin = fx.ue_preimage(env_action, sys_action,
                                   basin, t)
        rim = bdd.apply('diff', z, basin)
        rim = bdd.apply('and', rim, -cox_basin)
        # rho_2: pick persistence set
        s = "({c}' = {c}) & ({w} = {none})".format(
            c=c, w=w, none=n_holds)
        count = t.add_expr(s)
        u = bdd.apply('and', rim, count)
        v = bdd.false
        for i, y in enumerate(yi):
            s = "{w}' = {i}".format(w=w, i=i)
            count = t.add_expr(s)
            trans = _moore_trans(y, t)
            q = bdd.apply('and', count, trans)
            v = bdd.apply('or', v, q)
        u = bdd.apply('and', u, v)
        rho_2 = bdd.apply('or', rho_2, u)
        # rho_3: descent in recurrence basin
        s = (
            "({c}' = {c}) &"
            "({w} != {none}) &"
            "({w}' = {w})").format(
                c=c, w=w, none=n_holds)
        count = t.add_expr(s)
        u = bdd.apply('and', rim, count)
        v = bdd.false
        for i, xjr in enumerate(xijr):
            for j, (xr, goal) in enumerate(zip(xjr, goals)):
                s = (
                    "({c} = {j}) &"
                    " ({w} = {i})").format(c=c, w=w, i=i, j=j)
                count = t.add_expr(s)
                x_basin = xr[0]
                p = bdd.false
                for x in xr[1:]:
                    trans = _moore_trans(x_basin, t)
                    q = bdd.apply('and', trans, -x_basin)
                    q = bdd.apply('and', x, q)
                    p = bdd.apply('or', p, q)
                    x_basin = x
                p = bdd.apply('and', p, count)
                p = bdd.apply('and', p, -goal)
                v = bdd.apply('or', v, p)
        u = bdd.apply('and', u, v)
        rho_3 = bdd.apply('or', rho_3, u)
        # rho_4: advance to next recurrence goal
        u = bdd.false
        for j, goal in enumerate(goals):
            jp = (j + 1) % len(goals)
            s = "({c} = {j}) & ({c}' = {jp})".format(
                c=c, j=j, jp=jp)
            count = t.add_expr(s)
            p = bdd.apply('and', count, goal)
            u = bdd.apply('or', u, p)
        s = (
            "({w} != {none}) &"
            "({w}' = {w})").format(
                c=c, w=w, none=n_holds)
        count = t.add_expr(s)
        u = bdd.apply('and', u, count)
        u = bdd.apply('and', u, rim)
        v = bdd.false
        for i, y in enumerate(yi):
            s = "{w} = {i}".format(w=w, i=i)
            count = t.add_expr(s)
            trans = _moore_trans(y, t)
            q = bdd.apply('and', count, trans)
            v = bdd.apply('or', v, q)
        u = bdd.apply('and', u, v)
        rho_4 = bdd.apply('or', rho_4, u)
        # update
        basin = z
    # \rho
    u = bdd.apply('or', rho_1, rho_2)
    u = bdd.apply('or', rho_3, u)
    u = bdd.apply('or', rho_4, u)
    u = bdd.apply('and', sys_action, u)
    # counter limits
    u = bdd.apply('and', t.action['sys'][0], u)
    if not aut.plus_one:
        u = bdd.apply('->', env_action, u)
    if aut.moore:
        u = bdd.forall(t.upvars, u)
    t.action['sys'] = [u]
    # initial condition for counter
    s = '({c} = 0) & ({w} = {none})'.format(
        c=c, w=w, none=n_holds)
    count = t.add_expr(s)
    # \A init
    init = _all_init(env_init, sys_init, count, bdd)
    t.init['env'] = [init]
    init = bdd.apply('->', init, win_set)
    assert init == bdd.true, 'losing from some init states'
    return t


def _all_init(env_init, sys_init, internal, bdd):
    """Initial condition conjoining for all variables."""
    # case of `ONE_SIDE_INIT` with `ENVINIT` in `gr1c`
    u = bdd.apply('and', sys_init, internal)
    u = bdd.apply('and', env_init, u)
    return u


def _moore_trans(target, aut):
    """Return controllable transitions for progress."""
    bdd = aut.bdd
    env_action = aut.action['env'][0]
    sys_action = aut.action['sys'][0]
    uvars = aut.upvars
    u = bdd.rename(target, aut.prime)
    if aut.plus_one:
        u = bdd.apply('->', env_action, u)
        u = bdd.apply('and', sys_action, u)
    else:
        u = bdd.apply('and', sys_action, u)
        u = bdd.apply('->', env_action, u)
    if aut.moore:
        u = bdd.forall(uvars, u)
    return u


def trivial_winning_set(aut_streett):
    """Return set of trivially winning nodes for Streett(1).

    @return: `(trivial, aut_streett)` where:
        - `trivial`: node in `aut_streett.bdd`
        - `aut_streett`: `symbolic.Automaton`
    """
    aut_rabin = symbolic.Automaton()
    for var, d in aut_streett.vars.iteritems():
        d = d.copy()
        owner = d['owner']
        owner = 'env' if owner == 'sys' else 'sys'
        d['owner'] = owner
        aut_rabin.vars[var] = d
    aut_rabin.action['env'] = aut_streett.action['sys']
    aut_rabin.action['sys'] = aut_streett.action['env']
    win = ['!({w})'.format(w=w) for w in aut_streett.win['<>[]']]
    aut_rabin.win['[]<>'] = win
    symbolic.fill_blanks(aut_rabin, rabin=True)
    aut_rabin.bdd = aut_streett.bdd
    aut_streett = aut_streett.build()
    aut_rabin = aut_rabin.build()
    # solve
    win_streett, _, _ = solve_streett_game(aut_streett)
    zk, _, _ = solve_rabin_game(aut_rabin)
    win_rabin = zk[-1]
    # find trivial win set
    # win_rabin_ = _copy_bdd(win_rabin,
    #                        aut_rabin.bdd, aut_streett.bdd)
    trivial = aut_streett.bdd.apply('diff', win_streett, win_rabin)
    return trivial, aut_streett


def _map_nested_lists(f, x, *arg, **kw):
    """Recursively map lists, with non-lists at the bottom.

    Useful for applying `dd.bdd.copy_bdd` to several lists.
    """
    if isinstance(x, list):
        return [_map_nested_lists(f, y, *arg, **kw) for y in x]
    else:
        return f(x, *arg, **kw)


def _copy_bdd(u, a, b):
    """Copy bdd `u` from manager `a` to `b`.

    No effect if `a is b`.
    """
    if a is b:
        return u
    return a.copy(u, b)



import logging
import pprint
from dd import bdd as _bdd
from dd import mdd
from nose.tools import assert_raises
from omega.symbolic import symbolic, enumeration
from omega.games import gr1


log = logging.getLogger('astutils')
log.setLevel('ERROR')
log = logging.getLogger('omega')
log.setLevel('ERROR')


def test_streett_trivial_loop():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    symbolic.fill_blanks(a)
    with assert_raises(AssertionError):
        gr1.solve_streett_game(a)
    # solve
    a = a.build()
    assert len(a.win['<>[]']) == 1
    assert len(a.win['[]<>']) == 1
    z, yij, xijk = gr1.solve_streett_game(a)
    assert z == a.bdd.true, z
    # transducer
    t = gr1.make_streett_transducer(z, yij, xijk, a)
    assert isinstance(t, symbolic.Automaton), t
    # vars
    assert '_goal' in t.vars, t.vars
    vt = t.vars['_goal']['type']
    assert vt == 'int', vt
    dom = t.vars['_goal']['dom']
    assert dom == (0, 0), dom
    assert 'x' in t.vars, t.vars
    # init
    assert len(t.init['env']) == 1, t.init
    (init,) = t.init['env']
    init_ = t.add_expr('_goal = 0')
    assert init == init_, t.bdd.to_expr(init)
    # action
    assert len(t.action['sys']) == 1, t.action
    (action,) = t.action['sys']
    s = "(_goal = 0) & (_goal' = 0)"
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)
    # win
    assert len(t.win['[]<>']) == 0, t.win['[]<>']


def test_rabin_trivial_loop():
    a = symbolic.Automaton()
    a.acceptance = 'Rabin(1)'
    a.vars = dict(x=dict(type='bool', owner='sys'))
    symbolic.fill_blanks(a, rabin=True)
    with assert_raises(AssertionError):
        gr1.solve_rabin_game(a)
    # solve
    a = a.build()
    assert len(a.win['<>[]']) == 1
    assert len(a.win['[]<>']) == 1
    zk, yki, xkijr = gr1.solve_rabin_game(a)
    assert zk[-1] == a.bdd.true, zk
    # transducer
    t = gr1.make_rabin_transducer(zk, yki, xkijr, a)
    assert isinstance(t, symbolic.Automaton), t
    # vars
    assert '_goal' in t.vars, t.vars
    vt = t.vars['_goal']['type']
    assert vt == 'int', vt
    dom = t.vars['_goal']['dom']
    assert dom == (0, 0), dom
    assert '_hold' in t.vars, t.vars
    vt = t.vars['_hold']['type']
    assert vt == 'int', vt
    dom = t.vars['_hold']['dom']
    assert dom == (0, 1), dom
    assert 'x' in t.vars, t.vars
    # init
    assert len(t.init['env']) == 1, t.init
    (init,) = t.init['env']
    init_ = t.add_expr('(_goal = 0) & (_hold = 1)')
    assert init == init_, t.bdd.to_expr(init)
    # action
    assert len(t.action['sys']) == 1, t.action
    (action,) = t.action['sys']
    s = (
        "(_goal = 0) & (_goal' = 0) &"
        "(_hold' = 0)")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)
    # win
    w = t.win['[]<>']
    assert len(w) == 0, w
    w = t.win['<>[]']
    assert len(w) == 0, w


def test_streett_deadend():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='int', dom=(0, 10), owner='sys'))
    a.action['sys'] = ["(x = 0) & (x' = 5)"]
    symbolic.fill_blanks(a)
    # solve
    aut = a.build()
    z, _, _ = gr1.solve_streett_game(aut)
    win_set = z
    assert win_set == aut.bdd.false, win_set


def test_rabin_deadend():
    a = symbolic.Automaton()
    a.acceptance = 'Rabin(1)'
    a.vars = dict(x=dict(type='int', dom=(0, 101), owner='sys'))
    a.action['sys'] = ["(x = 1) & (x' = 96 - x)"]
    symbolic.fill_blanks(a)
    aut = a.build()
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.false, win_set


def test_streett_always_x():
    # always x
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.action['sys'] = ["x'"]
    symbolic.fill_blanks(a)
    # solve
    aut = a.build()
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.bdd.true, z
    # transducer
    t = gr1.make_streett_transducer(z, yij, xijk, aut)
    # init
    (init,) = t.init['env']
    init_ = t.add_expr('_goal = 0')
    assert init == init_, t.bdd.to_expr(init)
    # action
    (action,) = t.action['sys']
    s = (
        "(_goal = 0) & (_goal' = 0) & "
        "x' ")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)
    #
    # always ! x
    a.init['env'] = ['!x']
    a.action['sys'] = ['!x']
    aut = a.build()
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.add_expr('!x'), aut.bdd.to_expr(z)
    # tranducer
    t = gr1.make_streett_transducer(z, yij, xijk, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    init_ = t.add_expr('(_goal = 0) & !x')
    assert init == init_, t.bdd.to_expr(init)
    (action,) = t.action['sys']
    s = "(_goal = 0) & (_goal' = 0) & !x & !x'"
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)


def test_rabin_always_x():
    # always x
    a = symbolic.Automaton()
    a.acceptance = 'Rabin(1)'
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.action['sys'] = ["x'"]
    symbolic.fill_blanks(a, rabin=True)
    # solve
    aut = a.build()
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, win_set
    # transducer
    t = gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    # init
    (init,) = t.init['env']
    s = '(_goal = 0) & (_hold = 1)'
    init_ = t.add_expr(s)
    assert init == init_, t.bdd.to_expr(init)
    # action
    (action,) = t.action['sys']
    s = (
        "(_goal = 0) & (_goal' = 0) & "
        "(_hold' = 0) & "
        "x' ")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)
    #
    # always ! x
    a.init['env'] = ['!x']
    a.action['sys'] = ['!x']
    aut = a.build()
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    win_set_ = aut.add_expr('!x')
    assert win_set == win_set_, aut.bdd.to_expr(win_set)
    # transducer
    t = gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    s = (
        '(_goal = 0) & (_hold = 1) & '
        '!x')
    init_ = t.add_expr(s)
    assert init == init_, t.bdd.to_expr(init)
    (action,) = t.action['sys']
    s = (
        "(_goal = 0) & (_goal' = 0) & "
        "(_hold' = 0) &"
        "!x & !x'")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)


def test_streett_counter():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.action['sys'] = ["x -> !x'"]
    a.win['[]<>'] = ['x']
    symbolic.fill_blanks(a)
    # solve
    aut = a.build()
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.bdd.true, z
    # transducer
    t = gr1.make_streett_transducer(z, yij, xijk, aut)
    (init,) = t.init['env']
    init_ = t.add_expr('_goal = 0')
    assert init == init_, t.bdd.to_expr(init)
    (action,) = t.action['sys']
    assert action_refined(aut, t)
    # regression
    s = (
        "(_goal = 0) & (_goal' = 0) & "
        "(x <-> !x')")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)


def test_rabin_counter():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.action['sys'] = ["x -> !x'"]
    a.win['[]<>'] = ['x']
    symbolic.fill_blanks(a, rabin=True)
    # solve
    aut = a.build()
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, aut.bdd.to_expr(win_set)
    # transducer
    t = gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    s = '(_goal = 0) & (_hold = 1)'
    init_ = t.add_expr(s)
    assert init == init_, t.bdd.to_expr(init)
    (action,) = t.action['sys']
    # regression
    s = (
        "(_goal = 0) & (_goal' = 0) & "
        "(_hold' = 0) & "
        "ite(_hold = 0, x <-> !x', x -> !x')")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)


def test_rabin_persistence():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.init['sys'] = ['!x']
    a.win['<>[]'] = ['x']
    symbolic.fill_blanks(a, rabin=True)
    # solve
    aut = a.build()
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, aut.bdd.to_expr(win_set)
    # tranducer
    t = gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    s = '(_goal = 0) & (_hold = 1) & !x'
    init_ = t.add_expr(s)
    assert init == init_, t.bdd.to_expr(init)
    (action,) = t.action['sys']
    s = (
        "(_goal = 0) & (_goal' = 0) & "
        "ite(_hold' = 1, !x & x', x & x')")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)
    #
    # unrealizable
    a.win['[]<>'] = ['!x']
    aut = a.build()
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.false, aut.bdd.to_expr(win_set)


def test_rabin_persistence_2():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.init['sys'] = []
    a.win['<>[]'] = ['x', '!x']
    symbolic.fill_blanks(a, rabin=True)
    # solve
    aut = a.build()
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, aut.bdd.to_expr(win_set)
    # transducer
    t = gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    s = '(_goal = 0) & (_hold = 2)'
    init_ = t.add_expr(s)
    assert init == init_, t.bdd.to_expr(init)
    (action,) = t.action['sys']
    # enumeration.dump_relation(action, t)
    s = (
        "(_goal = 0) & (_goal' = 0) & "
        "ite(_hold = 1,"
        "(_hold' = 1) & !x',"
        "ite(_hold = 0,"
        "(_hold' = 0) & x',"
        "(_hold = 2) & ("
        "( (_hold' = 0) & x' ) | "
        "( (_hold' = 1) & !x' )"
        ") ) )")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)


def test_streett_with_safety_assumption():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='env'))
    a.action['env'] = ['x']
    a.action['sys'] = ['x']
    symbolic.fill_blanks(a)
    # solve
    aut = a.build()
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.bdd.true, z
    # transducer
    t = gr1.make_streett_transducer(z, yij, xijk, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    assert init == t.add_expr('_goal = 0'), t.bdd.to_expr(init)
    (action,) = t.action['sys']
    action_ = t.add_expr("(_goal = 0) & (_goal' = 0) & x")
    assert action == action_, t.bdd.to_expr(action)
    #
    # negate action to make unrealizable
    a.action['sys'] = ['!x']
    # solve
    aut = a.build()
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.add_expr('!x'), aut.bdd.to_expr(z)
    # transducer
    with assert_raises(AssertionError):
        gr1.make_streett_transducer(z, yij, xijk, aut)


def test_rabin_with_safety_assumption():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='env'))
    a.init['env'] = ['x']
    a.action['env'] = ["x'"]
    a.action['sys'] = ["x"]
    symbolic.fill_blanks(a, rabin=True)
    # solve
    aut = a.build()
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.add_expr('x'), win_set
    # transducer
    t = gr1.make_rabin_transducer(zk, yki, xkijr, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    s = 'x & (_goal = 0) & (_hold = 1)'
    init_ = t.add_expr(s)
    assert init == init_, t.bdd.to_expr(init)
    (action,) = t.action['sys']
    s = "((_goal = 0) & (_goal' = 0) & (_hold' = 0) & x)"
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)
    #
    # negate action to make unrealizable
    a.action['sys'] = ['!x']
    # solve
    aut = a.build()
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == t.bdd.false, t.bdd.to_expr(win_set)
    with assert_raises(AssertionError):
        gr1.make_rabin_transducer(zk, yki, xkijr, aut)


def test_streett_with_liveness_assumption():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='env'),
                  y=dict(type='int', dom=(0, 2), owner='sys'))
    a.init['env'] = ['y < 2']
    a.action['sys'] = [
        "((y = 0) & !x) -> (y' = 0)",
        "(y = 0) -> (y' < 2)",
        "(y = 1) -> (y' = 0)",
        "(y = 2) -> False"]
    a.win['<>[]'] = ['!x']
    a.win['[]<>'] = ['y = 1']
    symbolic.fill_blanks(a)
    aut = a.build()
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    z_ = aut.add_expr('y < 2')
    e = aut.bdd.to_expr(z)
    e_ = aut.bdd.to_expr(z_)
    assert z == z_, (e, e_)
    # transducer
    t = gr1.make_streett_transducer(z, yij, xijk, aut)
    assert action_refined(aut, t)
    (init,) = t.init['env']
    assert init == t.add_expr('(y < 2) & (_goal = 0)'), t.bdd.to_expr(init)
    (action,) = t.action['sys']
    s = (
        "( (y = 0) -> ite(x, (y' = 1), (y' = 0)) ) & "
        "( (y = 1) -> (y' = 0) ) & "
        "( (_goal = 0) & (_goal' = 0) ) & "
        "( (y != 2) & (y != 3) )")
    action_ = t.add_expr(s)
    sat = list(t.bdd.sat_iter(action))
    sys_action = aut.action['sys'][0]
    sys_action = _bdd.copy_bdd(sys_action, aut.bdd, t.bdd)
    u = t.bdd.apply('->', action, sys_action)
    assert u == t.bdd.true, u
    assert action == action_, (action, action_, pprint.pprint(sat))
    #
    # test complement
    b = symbolic.Automaton()
    b.acceptance = 'Rabin(1)'
    b.vars = dict(x=dict(type='bool', owner='sys'),
                  y=dict(type='int', dom=(0, 2), owner='env'))
    b.action['env'] = a.action['sys']
    b.win['<>[]'] = ['y != 1']
    b.win['[]<>'] = ['x']
    symbolic.fill_blanks(b, rabin=True)
    aut_rabin = b.build()
    zk, yki, xkijr = gr1.solve_rabin_game(aut_rabin)
    rabin_win_set = zk[-1]
    bdd = aut_rabin.bdd
    streett_win_set = _bdd.copy_bdd(z, aut.bdd, bdd)
    assert rabin_win_set == bdd.apply('not', streett_win_set)
    with assert_raises(AssertionError):
        gr1.make_rabin_transducer(zk, yki, xkijr, aut_rabin)


def test_streett_2_goals():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.win['[]<>'] = ['x', '!x']
    symbolic.fill_blanks(a)
    aut = a.build()
    # solve
    z, yij, xijk = gr1.solve_streett_game(aut)
    assert z == aut.bdd.true, aut.bdd.to_expr(z)
    # transducer
    t = gr1.make_streett_transducer(z, yij, xijk, aut)
    assert action_refined(aut, t)
    (action,) = t.action['sys']
    # print_fol_bdd(action, t.bdd, t.vars)
    # enumeration.dump_relation(action, t)
    s = (
        "((x & (_goal = 0)) -> (_goal' = 1)) & "
        "((!x & (_goal = 1)) -> (_goal' = 0)) & "
        "((!x & (_goal = 0)) -> (x' & (_goal' = 0))) & "
        "((x & (_goal = 1)) -> (!x' & (_goal' = 1)))")
    action_ = t.add_expr(s)
    assert action == action_, t.bdd.to_expr(action)


def test_rabin_goal():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='env'))
    a.win['[]<>'] = ['x']
    symbolic.fill_blanks(a, rabin=True)
    aut = a.build()
    # solve
    zk, yki, xkijr = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.false, aut.bdd.to_expr(win_set)


def test_rabin_2_goals():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='env'),
                  y=dict(type='bool', owner='sys'))
    a.win['[]<>'] = ['x -> y']
    symbolic.fill_blanks(a, rabin=True)
    aut = a.build()
    # solve
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, aut.bdd.to_expr(win_set)
    #
    # unrealizable
    a.win['<>[]'] = ['! y']
    aut = a.build()
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.false, aut.bdd.to_expr(win_set)
    #
    # realizable again
    a.win['<>[]'] = ['y']
    aut = a.build()
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.true, aut.bdd.to_expr(win_set)
    #
    # unrealizable
    a.win['[]<>'] = ['x -> y', '!y']
    aut = a.build()
    zk, _, _ = gr1.solve_rabin_game(aut)
    win_set = zk[-1]
    assert win_set == aut.bdd.false, aut.bdd.to_expr(win_set)


def test_trivial_winning_set():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='sys'))
    a.win['<>[]'] = ['!x']
    symbolic.fill_blanks(a)
    triv, aut = gr1.trivial_winning_set(a)
    assert triv == aut.bdd.true, aut.bdd.to_expr(triv)


def action_refined(aut_a, aut_b, owner='sys'):
    """Return `True` if action of `aut_b` refines `aut_a`.

    @type aut_a, aut_b: `symbolic.Automaton`
    """
    bdd_a = aut_a.bdd
    bdd_b = aut_b.bdd
    a = aut_a.action[owner][0]
    b = aut_b.action[owner][0]
    if bdd_a is not bdd_b:
        a_cp = _bdd.copy_bdd(a, bdd_a, bdd_b)
    else:
        a_cp = a
    refined = bdd_b.apply('->', b, a_cp)
    return refined == bdd_b.true


def print_fol_bdd(u, bdd, table):
    """Print a first-order formula for node `u`.

    Example:

    ```
    t = symbolic.Automaton()
    t.vars = dict(x=dict(type='int', dom=(0, 5), owner='sys'))
    t.action['sys'] = ['x = 1']
    u = t.action['sys'][0]
    print_fol_bdd(u, t.bdd, t.vars)
    ```
    """
    bdd.incref(u)
    dvars = symbolic._prime_and_order_table(table)
    m, umap = mdd.bdd_to_mdd(bdd, dvars)
    u_ = -umap[abs(u)]
    s = m.to_expr(u_)
    print(s)


if __name__ == '__main__':
    test_rabin_2_goals()



"""First-order wrapper for BDDs."""
# other places where relevant functions exist:
#   dd.mdd
#   omega.logic.bitvector
#   omega.symbolic.enumeration
#
# examples/interleaving
# symbolic_transducers
# simulate
import logging

try:
    from dd import cudd as _bdd
except ImportError:
    from dd import bdd as _bdd

from omega.logic import bitvector as bv
from omega.logic import syntax as stx
from omega.symbolic import bdd as sym_bdd
from omega.symbolic import enumeration as enum
from omega.symbolic import symbolic


log = logging.getLogger(__name__)


class Context(object):
    """First-order interface to a binary decision diagram."""

    def __init__(self):
        """Instantiate first-order context."""
        self.vars = dict()
        self.bdd = _bdd.BDD()

    def add_vars(self, dvars):
        """Bitblast and add variables in `dvars`.

        These variables can have integer or Boolean type hints.
        Type hints are used to choose a refinement of
        integers by finitely many bits.

        No type predicates managed here:
        This is an untyped logic!

        Priming is not reasoned about here.
        Priming is the responsibility of higher levels.
        """
        # should be fresh
        common = set(dvars).intersection(self.vars)
        assert not common, common
        t = bv.bitblast_table(dvars)
        self.vars.update(t)
        bits = bv.bit_table(t, t)
        for bit in bits:
            self.bdd.add_var(bit)

    def support(self, u):
        """Return FOL variables that `u` depends on."""
        supp = self.bdd.support(u)
        bit2int = bv.map_bits_to_integers(self.vars)
        return set(map(bit2int.__getitem__, supp))

    def replace(self, u, vars_to_new):
        """Return substitution of var names by values or vars.

        @param vars_to_new: `dict` that maps each var name to
            a var (as `str`), or to a value (as `bool` or `int`).
        """
        if not vars_to_new:
            return u
        assert vars_to_new, vars_to_new
        for k in vars_to_new:
            try:
                vars_to_new[k] + 'str'
                rename = True
            except TypeError:
                rename = False
            break
        if rename:
            bit_rename = _refine_renaming(vars_to_new, self.vars)
            return self.bdd.rename(u, bit_rename)
        else:
            bit_values = _refine_assignment(vars_to_new, self.vars)
            return self.bdd.cofactor(u, bit_values)

    def replace_with_bdd(self, u, var_subs):
        """Substitute Boolean-valued variables with BDD nodes."""
        # distinct from `replace` due to restriction to Boolean
        return self.bdd.compose(u, var_subs)

    def forall(self, qvars, u):
        """Universally quantify `qvars` in `u`."""
        return ~self.exist(qvars, ~u)

    def exist(self, qvars, u):
        """Existentially quantify `qvars` in `u`."""
        qbits = bv.bit_table(qvars, self.vars)
        return self.bdd.exist(qbits, u)

    def pick(self, u, full, care_vars):
        """Return a satisfying assignment, or `None`."""
        try:
            return next(self.sat_iter(u, full, care_vars))
        except StopIteration:
            return None

    def sat_iter(self, u, full, care_vars):
        """Return generator of first-order satisfying assignments."""
        care_bits = bv.bit_table(care_vars, self.vars)
        for bit_assignment in self.bdd.sat_iter(
                u, full=full, care_bits=care_bits):
            d = next(enum._bitfields_to_int_iter(
                bit_assignment, self.vars))
            yield d

    def add_expr(self, e):
        """Add first-order formula."""
        s = bv.bitblast(e, self.vars)
        return sym_bdd.add_expr(s, self.bdd)

    def apply(self, op, u, v=None, w=None):
        """Apply operator `op` on operands `u, v, w`."""
        return self.bdd.apply(op, u, v, w)


def reorder(dvars, fol):
    """Shift integers up in the variable order of `fol.bdd`."""
    bdd = fol.bdd
    for var, d in dvars.iteritems():
        level = d['level']
        dv = fol.vars[var]
        var_type = dv['type']
        if var_type == 'bool':
            bitnames = [var]
        else:
            assert var_type in ('int', 'saturating', 'modwrap')
            bitnames = dv['bitnames']
        # print('Shifting bits {bitnames} to '
        #       'level {level}'.format(
        #         level=level,
        #         bitnames=bitnames))
        # assert each bit goes up in level
        for bit in bitnames:
            old_level = bdd.level_of_var(bit)
            assert old_level >= level, (old_level, level)
        # shift to levels wanted
        order = [bdd.var_at_level(i)
                 for i in xrange(len(bdd.vars))]
        below = order[:level]
        above = order[level:]
        above = [v for v in above if v not in bitnames]
        new_order = below + bitnames + above
        dorder = {var: i for i, var in enumerate(new_order)}
        _bdd.reorder(bdd, dorder)
        order = [bdd.var_at_level(i)
                 for i in xrange(len(bdd.vars))]
        assert order == new_order, (order, new_order)


def _refine_assignment(fol_values, table):
    """Return assignment to bits, from FOL assignment."""
    bit_values = dict()
    for var, value in fol_values.iteritems():
        assert var in table, var
        if table[var]['type'] == 'bool':
            bit_values[var] = value
            continue
        d = _int_to_bit_assignment(var, value, table)
        bit_values.update(d)
    return bit_values


def _refine_renaming(fol_rename, table):
    """Return renaming of bits, from renaming of FOL vars."""
    bit_rename = dict()
    for old, new in fol_rename.iteritems():
        old_d = table[old]
        new_d = table[new]
        old_type = old_d['type']
        new_type = new_d['type']
        assert old_type == new_type, (old_type, new_type)
        if old_type == 'bool':
            bit_rename[old] = new
            continue
        assert old_type in ('int', 'saturating', 'modwrap')
        # int
        old_dom = old_d['dom']
        new_dom = new_d['dom']
        assert old_dom == new_dom, (old_dom, new_dom)
        old_bits = old_d['bitnames']
        new_bits = new_d['bitnames']
        assert len(old_bits) == len(new_bits), (
            old_bits, new_bits)
        # no overlap (otherwise use `compose`)
        common = set(old_bits).intersection(new_bits)
        assert not common, (
            common, old_bits, new_bits)
        # bit substitution
        # assume same bit numbering (index 0 is LSB)
        bit_rename.update(zip(old_bits, new_bits))
    return bit_rename


def _int_to_bit_assignment(var, value, table):
    """Return assignment to bits from assignment to integer."""
    assert var in table, var
    var_bits = bv.var_to_twos_complement(var, table)
    int_bits = bv.int_to_twos_complement(value)
    p, q = bv.equalize_width(var_bits, int_bits)
    values = dict()
    for u, v in zip(p, q):
        # primed ?
        if u.isdigit():
            assert u == v, (u, v)
        else:
            values[u] = bool(int(v))
    return values


def _prime_bits_of_integers(ints, t):
    """Return bit priming for integers in `x`."""
    bit_rename = dict()
    for i in ints:
        bits = t.vars[i]['bitnames']
        d = {k: stx.prime(k) for k in bits}
        bit_rename.update(d)
    return bit_rename


def closed_interval(var, a, b):
    """Return string `a <= var <= b`."""
    return (
        '({a} <= {var}) & '
        '({var} <= {b})').format(
            var=var, a=a, b=b)


def add_one_hot(var, a, b):
    """Return symbol table for one-hot encoding."""
    t = dict()
    for i in xrange(a, b):
        var = '{var}{i}'.format(var=var, i=i)
        d = dict(type='bool', owner='parameters', level=0)
        t[var] = d
    return t


def array_to_logic(a, counter, aut):
    """Return logic formula for Boolean array `a`.

    No array bounds produced.
    Instead, if `(counter < 0) or (counter >= len(a) - 1)`,
    then select `a[-1]`.
    Therefore, array bounds *must* be added separately.

    @param a: `list` of elements
    @param counter: name of index variable
    """
    bdd = aut.bdd
    r = a[-1]
    for i, x in enumerate(a[:-1]):
        s = '{counter} = {i}'.format(counter=counter, i=i)
        u = aut.add_expr(s)
        r = bdd.ite(u, x, r)
    return r


def multiplexer(a, bits, aut):
    """Return BDD node for selection of elements from `a`.

    The resulting expression is:

    ```
    ite(bits[-1], a[-1],
        ite(bits[-2], a[-2],
            ...
            ite(bits[1], a[1],
                ite(bits[0], a[0], FALSE)
    ```

    This shows that if i < j,
    then b[i] has lower priority than b[j].
    So, this is not exactly a multiplexer.
    """
    assert len(a) == len(bits), (a, bits)
    bdd = aut.bdd
    r = bdd.false
    for bit, x in zip(bits, a):
        g = aut.add_expr(bit)
        r = bdd.ite(g, x, r)
    return r



import logging
import dd.bdd
from omega.logic import bitvector as bv
from omega.symbolic import bdd as sym_bdd
from omega.symbolic import bdd_iterative as bdd_trs
from omega.symbolic import symbolic


log = logging.getLogger('astutils')
log.setLevel('ERROR')
log = logging.getLogger('omega')
log.setLevel('ERROR')


def test_partition_vars():
    bits = ['a', 'b', 'c']
    ubits = ['a']
    ebits = ['b', 'c']
    order = symbolic._add_primed_bits(bits)
    order_ = ["a", "a'", "b", "b'", "c", "c'"]
    assert order == order_, (order, order_)
    prime, partition = symbolic._partition_vars(bits, ubits, ebits)
    prime_ = {k: k + "'" for k in bits}
    partition_ = dict(
        uvars={"a"},
        upvars={"a'"},
        evars={"b", "c"},
        epvars={"b'", "c'"})
    assert prime == prime_, (prime, prime_)
    assert partition == partition_, (partition, partition_)


def test_slugsin_parser():
    parser = sym_bdd.Parser()
    s = '! x'
    u = parser.parse(s)
    assert_is_operator(u, '!')
    (x,) = u.operands
    assert_is_var(x, 'x')
    s = '& x y'
    u = parser.parse(s)
    assert_is_operator(u, '&')
    x, y = u.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '| x y'
    u = parser.parse(s)
    assert_is_operator(u, '|')
    x, y = u.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '^ x12 y0'
    u = parser.parse(s)
    assert_is_operator(u, '^')
    x12, y0 = u.operands
    assert_is_var(x12, 'x12')
    assert_is_var(y0, 'y0')
    s = '$1 0'
    u = parser.parse(s)
    assert_is_buffer(u, 1)
    (a,) = u.memory
    assert_is_number(a, 0)
    s = '$2 0 1'
    u = parser.parse(s)
    assert_is_buffer(u, 2)
    a, b = u.memory
    assert_is_number(a, 0)
    assert_is_number(b, 1)
    s = '$2 $2 1 ?0 ?0'
    u = parser.parse(s)
    assert_is_buffer(u, 2)
    a, b = u.memory
    assert_is_register(b, 0)
    assert_is_buffer(a, 2)
    a, b = a.memory
    assert_is_number(a, 1)
    assert_is_register(b, 0)
    s = '\E x & x y'
    u = parser.parse(s)
    assert_is_operator(u, '\E')
    qvars, e = u.operands
    assert_is_var(qvars, 'x')
    assert_is_operator(e, '&')
    x, y = e.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '\A y | y ! x4'
    u = parser.parse(s)
    assert_is_operator(u, '\A')
    qvars, e = u.operands
    assert_is_var(qvars, 'y')
    assert_is_operator(e, '|')
    y, not_x4 = e.operands
    assert_is_var(y, 'y')
    assert_is_operator(not_x4, '!')
    (x4,) = not_x4.operands
    assert_is_var(x4, 'x4')
    s = '\E x \A y | x y'
    u = parser.parse(s)
    assert_is_operator(u, '\E')
    qvars, e = u.operands
    assert_is_var(qvars, 'x')
    assert_is_operator(e, '\A')
    qvars, e = e.operands
    assert_is_var(qvars, 'y')
    assert_is_operator(e, '|')
    x, y = e.operands
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    s = '\S $4 x y  z w  & y w'
    u = parser.parse(s)
    assert_is_operator(u, '\S')
    subs, e = u.operands
    assert_is_buffer(subs, 4)
    x, y, z, w = subs.memory
    assert_is_var(x, 'x')
    assert_is_var(y, 'y')
    assert_is_var(z, 'z')
    assert_is_var(w, 'w')
    assert_is_operator(e, '&')
    y, w = e.operands
    assert_is_var(y, 'y')
    assert_is_var(w, 'w')


def assert_is_operator(u, operator):
    assert hasattr(u, 'type'), u
    assert u.type == 'operator', u.type
    assert u.operator == operator


def assert_is_var(u, var):
    assert hasattr(u, 'type'), u
    assert u.type == 'var', u.type
    assert u.value == var, u.value


def assert_is_number(u, j):
    assert hasattr(u, 'type'), u
    assert u.type == 'num', u.type
    assert u.value == str(j), u.value


def assert_is_buffer(u, n):
    assert hasattr(u, 'type'), u
    assert u.type == 'buffer', u.type
    assert hasattr(u, 'memory'), u
    assert len(u.memory) == n, u.memory


def assert_is_register(u, idx):
    assert hasattr(u, 'type'), u
    assert u.type == 'register', u.type
    assert u.value == str(idx), u.value


def test_iterative_bddizer():
    add = bdd_trs.add_expr
    order = {'x': 0, 'y': 1, 'z': 2}
    # x & y
    bdd = dd.bdd.BDD(order)
    e = '& x y'
    u = add(e, bdd)
    v = bdd.add_expr('x & y')
    assert u == v, (u, v)
    # buffers
    # (x & y) | ! z
    bdd = dd.bdd.BDD(order)
    e = '$ 3   & x y   ! z  | ? 0 ? 1'
    u = add(e, bdd)
    v = bdd.add_expr('(x & y) | ! z')
    assert u == v, (u, v)


def test_bddizer_propositional():
    add = sym_bdd.add_expr
    order = {'x': 0, 'y': 1, 'z': 2}
    # x & y
    bdd = dd.bdd.BDD(order)
    e = '& x y'
    u = add(e, bdd)
    v = bdd.add_expr('x & y')
    assert u == v, (u, v)
    # buffers
    # (x & y) | ! z
    bdd = dd.bdd.BDD(order)
    e = '$ 3   & x y   ! z  | ? 0 ? 1'
    u = add(e, bdd)
    v = bdd.add_expr('(x & y) | ! z')
    assert u == v, (u, v)
    #
    e = '& $2 & 1 x ?0 $3 | !x y z & ?1 z'
    s = 'x & z'
    u = add(e, bdd)
    v = bdd.add_expr(s)
    assert u == v, (u, v)


def test_bddizer_quantifiers():
    add = sym_bdd.add_expr
    order = {'x': 0, 'y': 1, 'z': 2}
    bdd = dd.bdd.BDD(order)
    e = '\E x 1'
    u = add(e, bdd)
    assert u == bdd.true, u
    e = '\E x 0'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\A x 1'
    u = add(e, bdd)
    assert u == bdd.true, u
    e = '\A x 0'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\A x x'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\E x x'
    u = add(e, bdd)
    assert u == bdd.true, u
    e = '\A & x y y'
    u = add(e, bdd)
    assert u == bdd.false, u
    e = '\A x y'
    u = add(e, bdd)
    u_ = bdd.var('y')
    assert u == u_, (u, u_)
    e = '\E & x y & y x'
    u = add(e, bdd)
    assert u == bdd.true, u


def test_bddizer_substitution():
    add = sym_bdd.add_expr
    order = {'x': 0, 'y': 1, 'z': 2, 'w': 3}
    bdd = dd.bdd.BDD(order)
    e = '\S $2 x y y'
    u = add(e, bdd)
    u_ = bdd.var('x')
    assert u == u_, (u, u_)
    e = '\S $4 x y  z w  | y ! w'
    u = add(e, bdd)
    u_ = bdd.add_expr('x | ! z')
    assert u == u_, (u, u_)
    e = '$2 0 \S $2 x y y'
    u = add(e, bdd)
    u_ = bdd.var('x')
    assert u == u_, (u, u_)


def slugsin_parser(s, t):
    """Helper that converts prefix to infix syntax for readability."""
    slugs_table = t.to_slugs()
    ltl_parser = bv.Parser()
    p = ltl_parser.parse(s)
    bv.add_bitnames(slugs_table)
    s = p.flatten(t=slugs_table)
    slugsin_parser = sym_bdd.Parser()
    print(slugsin_parser.parse(s))



"""Convert relation predicate to predicate-action diagram.

Reference
=========

Leslie Lamport
    "TLA in pictures"
    IEEE Transactions on Software Engineering
    Vol.21, No.9, pp.768--775, 1995
    doi: 10.1109/32.464544
"""
import logging

import networkx as nx

from omega.logic import syntax as stx
from omega.symbolic import fol as _fol
from omega.symbolic import symbolic


log = logging.getLogger(__name__)


def action_to_steps(aut, qinit=None):
    r"""Return enumerated graph with steps as edges.

    @param qinit:
        '\A \A' == forall env values: forall sys values
        '\A \E' == forall env values: exist sys values
        '\E \A' == exist sys values: forall env values
    """
    if qinit is None:
        qinit = '\A \A'
    bdd = aut.bdd
    fol = _fol.Context()
    fol.bdd = bdd
    table = symbolic._prime_and_order_table(aut.vars)
    # fol.add_vars(table)
    fol.vars = table
    aut.control = dict(env=set(), sys=set())
    aut.primed_vars = dict(env=set(), sys=set())
    for var, d in aut.vars.iteritems():
        pvar = stx.prime(var)
        owner = d['owner']
        assert owner in aut.players, owner
        aut.control[owner].add(var)
        aut.primed_vars[owner].add(pvar)
    # prime_vars = {var: stx.prime(var) for var in aut.vars}
    unprime_vars = {stx.prime(var): var for var in aut.vars}
    keys = list(aut.vars)  # fix an order for tupling
    umap = dict()  # map assignments -> node numbers
    g = nx.DiGraph()
    g.sorted_vars = keys
    # danger of blowup due to sparsity
    # implement enumerated equivalent to compare
    if qinit == '\A \E':
        queue, visited = _forall_exist_init(g, fol, aut, umap, keys)
    elif qinit == '\A \A':
        queue, visited = _forall_init(g, fol, aut, umap, keys)
    else:
        raise Exception('unknown qinit "{q}"'.format(q=qinit))
    log.info('{n} initial nodes'.format(n=len(queue)))
    # search
    while queue:
        node = queue.pop()
        values = g.node[node]
        log.debug('at node: {d}'.format(d=values))
        assert set(values) == set(aut.vars), (values, aut.vars)
        (u,) = aut.action['env']
        u = fol.replace(u, values)
        # apply Mealy controller function
        env_iter = fol.sat_iter(
            u, full=True, care_vars=aut.primed_vars['env'])
        (u,) = aut.action['sys']
        sys = fol.replace(u, values)
        for next_env in env_iter:
            u = fol.replace(sys, next_env)
            u = fol.replace(u, unprime_vars)
            env_values = {unprime_vars[var]: value
                          for var, value in next_env.iteritems()}
            v = fol.replace(visited, env_values)
            # prefer already visited nodes
            v = bdd.apply('and', u, v)
            if v == bdd.false:
                log.info('cannot remain in visited nodes')
                v = u
                remain = False
            else:
                remain = True
            sys_values = fol.pick(
                v, full=True, care_vars=aut.control['sys'])
            d = dict(env_values)
            d.update(sys_values)
            # assert
            u = fol.replace(visited, d)
            assert u == bdd.true or u == bdd.false
            assert remain == (u == bdd.true), remain
            # find or add node
            if remain:
                next_node = _find_node(d, umap, keys)
            else:
                next_node = _add_new_node(d, g, queue, umap, keys)
                visited = _add_to_visited(d, visited, aut)
            g.add_edge(node, next_node)
            log.debug((
                'next env: {e}\n'
                'next sys: {s}\n').format(
                    e=env_values,
                    s=sys_values))
    return g


def _forall_init(g, fol, aut, umap, keys):
    r"""Enumerate initial states with \A \A vars."""
    bdd = fol.bdd
    init = bdd.apply(
        'and', aut.init['env'][0], aut.init['sys'][0])
    init_iter = fol.sat_iter(
        init, full=True,
        care_vars=aut.vars)
    visited = bdd.false
    queue = list()
    for d in init_iter:
        _add_new_node(d, g, queue, umap, keys)
        visited = _add_to_visited(d, visited, aut)
    return queue, visited


def _forall_exist_init(g, fol, aut, umap, keys):
    r"""Enumerate initial states with \A env \E sys vars.

    Note that each initial "state" is a class of
    initial states in ZF set theory.
    """
    bdd = fol.bdd
    u = fol.exist(aut.control['sys'], aut.init['env'][0])
    env_iter = fol.sat_iter(
        u, full=True,
        care_vars=aut.control['env'])
    init = bdd.apply('and', aut.init['sys'][0], u)
    visited = bdd.false
    queue = list()
    for env_0 in env_iter:
        u = fol.replace(init, env_0)
        sys_iter = fol.sat_iter(
            u, full=True,
            care_vars=aut.control['sys'])
        for sys_0 in sys_iter:
            d = dict(env_0)
            d.update(sys_0)
            _add_new_node(d, g, queue, umap, keys)
            visited = _add_to_visited(d, visited, aut)
            break
    return queue, visited


def _find_node(d, umap, keys):
    """Return node in `umap` with assignment `d`."""
    key = tuple(d[k] for k in keys)
    assert key in umap, (key, umap)
    u = umap[key]
    return u


def _add_new_node(d, g, queue, umap, keys):
    """Add node to graph `g` for the assignment `d`."""
    u = len(g)
    assert u not in g, u
    g.add_node(u, **d)
    key = tuple(d[k] for k in keys)
    assert key not in umap, (key, umap)
    umap[key] = u
    queue.append(u)
    log.debug(d)
    return u


def _add_to_visited(values, visited, aut):
    """Return BDD `visted` updated with assignment `values`."""
    bdd = aut.bdd
    c = list()
    for var, value in values.iteritems():
        if aut.vars[var]['type'] == 'bool':
            assert value in (True, False), value
            if bool(value):
                c.append(var)
            else:
                c.append('! ' + var)
            continue
        # integer
        s = '{var} = {value}'.format(var=var, value=value)
        c.append(s)
    s = stx.conj(c)
    u = aut.add_expr(s)
    visited = bdd.apply('or', visited, u)
    return visited



"""Symbolic automaton.

Including bitblasting of first-order to bitvector logic,
and transformation to BDD nodes.
"""
import copy
import logging
import pprint

import dd.bdd
import natsort

from omega.logic import bitvector as bv
from omega.logic import syntax as stx
from omega.symbolic import bdd as sym_bdd


logger = logging.getLogger(__name__)


class Automaton(object):
    r"""Transition relation, initial condition, and hidden variables.

    User-defined attributes
    =======================

    You define the attributes:

      - vars: `dict` that maps each variable name to
        a `dict` with keys:

          - `"type": "bool" or "modwrap" or "saturating"`
          - `"dom": (min, max)`, range of integer
          - `"owner": "env" or "sys"`
          - `"init"`: has suitable `__str__` (and is optional)

      - init: initial condition
      - action: transition relation
      - win: winning condition
      - acceptance: `str` that describes `win`,
          for example `'Streett(1)'`
      - moore: choose between Moore or Mealy implementation
      - plus_one: strict implication with `\weakprevious`

    Each of `init, action` is a `dict`:

      `{player: list(), ...}`

    `win` is a `dict` of the form:

      `{'<>[]': list(),
        '[]<>': list()}`

    Add formulae (as strings) to these lists.
    Call the method `build` to convert these formulae to BDD nodes,
    and generate the below.


    Auto-generated attributes
    =========================

    The method `build` adds to `vars[var]` the keys:

      - `"bitnames"`: `list`
      - `"signed"`: `True` if signed integer
      - `"width"`: `len(bitnames)`

    Each of the following is a `set` of variable names:

      - uvars
      - upvars
      - ubvars
      - evars
      - epvars
      - ebvars
      - uevars
      - uepvars
      - prime: `dict` that maps each unprimed to a primed var
      - unprime: `dict` that maps eacn primed to an unprimed var

    Prefix meaning:

      u = universally quantified
      e = existentially quantified
      p = primed (absence means unprimed)
      b = both primed and unprimed


    References
    ==========

    Leslie Lamport
        "The temporal logic of actions"
        ACM Transactions on Programming
        Languages and Systems (TOPLAS),
        Vol.16, No.3, pp.872--923, 1994

    Rajeev Alur, Thomas A. Henzinger, Orna Kupferman
        "Alternating-time temporal logic"
        Journal of the ACM (JACM)
        Vol.49, No.5, pp.672--713, 2002

    Klaus Schneider
        "Verification of reactive systems"
        Springer, 2004
    """

    def __init__(self):
        self.players = {'env': 0, 'sys': 1}  # name -> turn
        self.vars = dict()
        self.moore = True
        self.plus_one = True
        # formulae
        self.init = dict(env=list(), sys=list())
        self.action = dict(env=list(), sys=list())
        self.win = {'<>[]': list(), '[]<>': list()}
        self.acceptance = 'Streett(1)'
        # auto-populated
        self.turns = list()  # inverse of `self.players`
        # subsets of bits
        self.uvars = set()  # unprimed env
        self.upvars = set()  # primed env
        self.ubvars = set()  # both primed and unprimed env
        self.evars = set()  # unprimed sys
        self.epvars = set()  # primed sys
        self.ebvars = set()  # both primed and unprimed env
        self.uevars = set()  # all unprimed vars
        self.uepvars = set()  # all primed vars
        # future: hidden, node vars
        # map between primed and unprimed
        self.prime = dict()  # unprimed -> primed
        self.unprime = dict()  # primed -> unprimed
        # aux
        # init only to aid static analysis
        self.bdd = dd.bdd.BDD()

    def __copy__(self):
        a = Automaton()
        a.players = copy.deepcopy(self.players)
        a.vars = copy.deepcopy(self.vars)
        a.init = copy.deepcopy(self.init)
        a.action = copy.deepcopy(self.action)
        a.win = copy.deepcopy(self.win)
        a.acceptance = self.acceptance
        a.bdd = self.bdd
        return a

    def __del__(self):
        del self.init
        del self.action
        del self.win

    def __str__(self):
        c = [
            'Symbolic automaton:',
            40 * '=',
            '',
            'variables:',
            '{dvars}',
            '',
            'Moore' if self.moore else 'Mealy',
            ('causal' if self.plus_one else
            'circular') + ' implication']
        if self.init['env']:
            c.extend([
                'ENV INIT:',
                '{env_init}',
                ''])
        if self.action['env']:
            c.extend([
                'ENV ACTION:',
                '{env_action}',
                ''])
        if self.init['sys']:
            c.extend([
                'SYS INIT:',
                '{sys_init}',
                ''])
        if self.action['sys']:
            c.extend([
                'SYS ACTION:',
                '{sys_action}',
                ''])
        if self.win['<>[]']:
            c.extend([
                '<>[] (persistence):',
                '{fg}',
                ''])
        if self.win['[]<>']:
            c.extend([
                '[]<> (recurrence):',
                '{gf}',
                ''])
        return '\n'.join(c).format(
            dvars=pprint.pformat(self.vars),
            env_init=_join(self.init['env']),
            env_action=_join(self.action['env']),
            fg=_join(self.win['<>[]']),
            sys_init=_join(self.init['sys']),
            sys_action=_join(self.action['sys']),
            gf=_join(self.win['[]<>']))

    def dumps(self):
        """Return `repr` of a `dict` with the attributes."""
        d = dict(vars=self.vars, init=self.init,
                 action=self.action, win=self.win)
        return repr(d)

    def build(self):
        """Return `Automaton` with formulae as BDD nodes.

        Bitblast variables and formulae,
        add bits to `self.bdd`,
        and populate primed and unprimed bits.

        BDD levels defined only if no existing bits.
        Otherwise, reorder to desired order later.
        """
        aut = _bitblast(self)
        aut.bdd = self.bdd
        aut = _bitvector_to_bdd(aut)
        aut.moore = self.moore
        aut.plus_one = self.plus_one
        return aut

    def assert_consistent(self, built=False):
        """Raise `AssertionError` if not well-formed.

        @param built: if `True`,
        then assert `build` has been called.
        """
        # check attributes
        for d in (self.init, self.action, self.win):
            for k, v in d.iteritems():
                for u in v:
                    if built:
                        assert u in self.bdd, u
                    else:
                        assert isinstance(u, basestring)
        if not built:
            return
        # bdd initialized ?
        assert self.bdd.vars is not None
        if self.vars:
            assert self.uevars

    def update(self, attr, d):
        """Add formulae from `dict` `d`."""
        r = getattr(self, attr)
        if attr in ('init', 'action'):
            r['env'].extend(d['env'])
            r['sys'].extend(d['sys'])
        elif attr == 'win':
            r['<>[]'].extend(d['<>[]'])
            r['[]<>'].extend(d['[]<>'])
        else:
            raise Exception('unknown attr "{a}"'.format(a=attr))

    def conjoin(self, as_what):
        """Conjoin attributes.

        @param as_what: `'bdd' or 'prefix' or 'infix'`
        """
        _conj_owner(self, 'env', as_what)
        _conj_owner(self, 'sys', as_what)

    def add_expr(self, e):
        """Add first-order formula."""
        # the bitblaster understands priming
        # the prefix parser treats primed bits as fresh
        # so suffice it for t to contain unprimed only
        t = self.vars
        s = bv.bitblast(e, t)
        u = sym_bdd.add_expr(s, self.bdd)
        return u


def fill_blanks(aut, as_bdd=False, rabin=False):
    """Add `"True"` to empty attributes `init`, `action`, `win`.

    @param as_bdd: if `True`, then represent `"True"` as `1`
    """
    if as_bdd:
        true = aut.bdd.true
        false = aut.bdd.false
    else:
        true = 'True'
        false = 'False'
    for d in (aut.init, aut.action):
        for k, v in d.iteritems():
            if not v:
                d[k] = [true]
    if not aut.win['<>[]']:
        if rabin:
            aut.win['<>[]'] = [true]
        else:
            aut.win['<>[]'] = [false]
    if not aut.win['[]<>']:
        aut.win['[]<>'] = [true]
    # post-condition
    for d in (aut.init, aut.action, aut.win):
        for k in d:
            assert len(d[k]) > 0


def _bitblast(aut):
    """Return `Automaton` with bitvector formulae.

    For each integer, the corresponding list of
    bitnames is added to the symbol table (attr `vars`).

    @type aut: `Automaton`
    """
    players = dict(aut.players)
    aut = copy.copy(aut)
    t = bv.bitblast_table(aut.vars)
    init, action = bv.type_invariants(t)
    for var, c in init.iteritems():
        owner = aut.vars[var]['owner']
        # collect type invariants of parameters too,
        # for convenience later
        aut.init.setdefault(owner, list())
        aut.init[owner].extend(c)
    for var, c in action.iteritems():
        owner = aut.vars[var]['owner']
        aut.action.setdefault(owner, list())
        aut.action[owner].extend(c)
    # conjoin to avoid it later over BDD nodes
    _conj_owner(aut, 'env', 'infix')
    _conj_owner(aut, 'sys', 'infix')
    a = Automaton()
    a.players = players
    a.vars = t
    # TODO: replace GR(1) syntax check
    # spec.check_syntax()
    _bitblast_owner(aut, a, 'env', t)
    _bitblast_owner(aut, a, 'sys', t)
    return a


def _conj_owner(aut, owner, as_what):
    """Conjoin the lists in the attributes of `owner`.

    @param as_what: `'bdd' or 'prefix' or 'infix'`
    """
    # get
    init = aut.init[owner]
    action = aut.action[owner]
    # compute
    if as_what == 'prefix':
        init = stx.conj_prefix(init)
        action = stx.conj_prefix(action)
    elif as_what == 'infix':
        init = stx.conj(init)
        action = stx.conj(action)
    elif as_what == 'bdd':
        def f(x, y):
            return aut.bdd.apply('and', x, y)
        init = stx.recurse_binary(f, init)
        action = stx.recurse_binary(f, action)
    else:
        raise Exception(
            'unknown as_what="{s}"'.format(s=as_what))
    # set
    aut.init[owner] = [init]
    aut.action[owner] = [action]


def _bitblast_owner(aut, a, owner, t):
    """Bitblast init, action, win of `owner`."""
    def f(x):
        return bv.bitblast(x, t)
    assert owner in ('env', 'sys'), owner
    a.init[owner] = map(f, aut.init[owner])
    a.action[owner] = map(f, aut.action[owner])
    if owner == 'env':
        s = '<>[]'
    else:
        s = '[]<>'
    a.win[s] = map(f, aut.win[s])


def _bitvector_to_bdd(aut):
    """Return `Automaton` with BDD formulae.

    @type aut: `Automaton`
    """
    players = dict(aut.players)
    table = aut.vars
    bits = bv.bit_table(table, table)
    # index both, to allow for unquantified parameters
    ubits = set(b for b, d in bits.iteritems()
                if d['owner'] == 'env')
    ebits = set(b for b, d in bits.iteritems()
                if d['owner'] == 'sys')
    b = _pick_var_order(bits, ubits)
    b = _add_primed_bits(b)
    bdd = aut.bdd
    # define levels only if no existing vars
    if bdd.vars:
        for var in b:
            bdd.add_var(var)
    else:
        for level, var in enumerate(b):
            bdd.add_var(var, level)
    # bundle as:
    a = Automaton()
    a.vars = copy.deepcopy(table)
    a.players = players
    a.bdd = bdd
    # slugsin -> BDD
    # add long attributes first,
    # to improve  effectiveness of reordering.
    # better if each attr is one formula.
    from_sections = _make_section_map(aut)
    to_sections = _make_section_map(a)
    lengths = {
        k: _section_len(v)
        for k, v in from_sections.iteritems()}
    sort = sorted(
        from_sections,
        key=lengths.__getitem__,
        reverse=True)
    for section in sort:
        p = from_sections[section]
        q = to_sections[section]
        _to_bdd(p, q, bdd)
    # vars
    player_vars = {k for k, d in table.iteritems()
                   if d['owner'] in players}
    if not player_vars:
        print('Warning: no player variables.')
        return a
    bits = bv.bit_table(player_vars, table)
    prime, partition = _partition_vars(bits, ubits, ebits)
    a.uvars = partition['uvars']
    a.upvars = partition['upvars']
    a.evars = partition['evars']
    a.epvars = partition['epvars']
    a.ubvars = set(a.uvars).union(a.upvars)
    a.ebvars = set(a.evars).union(a.epvars)
    a.uevars = set(a.uvars).union(a.evars)
    a.uepvars = set(a.upvars).union(a.epvars)
    # priming
    a.prime = prime
    a.unprime = {v: k for k, v in prime.iteritems()}
    return a


def _make_section_map(aut):
    """Return `dict` of `Automaton` attributes."""
    sections = dict(
        env_init=aut.init['env'],
        env_action=aut.action['env'],
        sys_init=aut.init['sys'],
        sys_action=aut.action['sys'])
    sections['<>[]'] = aut.win['<>[]']
    sections['[]<>'] = aut.win['[]<>']
    return sections


def _section_len(formulae):
    """Return sum of `len` of `str` in `formulae`."""
    return sum(len(s) for s in formulae)


def _to_bdd(a, b, bdd):
    """For each element of `a`, append a `bdd` node to `b`."""
    for s in a:
        u = sym_bdd.add_expr(s, bdd)
        b.append(u)


def _pick_var_order_eu(bits, ubits):
    ebits = set(bits).difference(ubits)
    u = natsort.natsorted(ubits)
    e = natsort.natsorted(ebits)
    return e + u


def _pick_var_order(bits, ubits):
    """Sort the `bits` in natural order.

    Two separate orders are constructed:

        - for `ubits`
        - for the rest of `bits`

    and concatenated in that order.
    """
    array_bits = {b for b, d in bits.iteritems() if d.get('array')}
    other = set(bits).difference(array_bits)
    top = natsort.natsorted(other)
    bottom = natsort.natsorted(array_bits)
    r = top + bottom
    return r


def _add_primed_bits(unprimed_bits):
    """Return list of ordered primed and unprimed bits."""
    bits = list()
    for bit in unprimed_bits:
        primed = stx.prime(bit)
        bits.append(bit)
        bits.append(primed)
    return bits


def _partition_vars(bits, ubits, ebits):
    """Return primed and unprimed variable names.

    @param bits: `list` of variable names as `str`
    @param ubits: universally quantified variables
    @param ebits: existentially quantified variables

    @return: (prime, partition)
      - prime: `dict` that maps unprimed to primed variable names
      - partition: `dict(uvars=set, upvars=set,
                         evars=set, epvars=set)`
    """
    prime = {b: stx.prime(b) for b in bits}
    partition = dict(
        uvars=set(ubits),
        upvars={prime[b] for b in ubits},
        evars=set(ebits),
        epvars={prime[b] for b in ebits})
    return prime, partition


def _prime_and_order_table(t):
    """Return ordered table of primed and unprimed variables.

    The table maps each (integer or Boolean) variable to
    a `dict` of attributes (same as `t`).
    The attributes include `"bitnames"`.

    The following attributes are added:

      - level: index among ordered prime and unprimed integers
      - len: number of values the variable can take

    @param t: table of unprimed variables as `str`
    @type t: `dict`

    @rtype: `dict` from `str` to `dict`
    """
    ordvars = natsort.natsorted(t)
    dvars = dict()
    for i, var in enumerate(ordvars):
        d = t[var]
        j = 2 * i
        dtype = d['type']
        if dtype in ('int', 'saturating'):
            bits = d['bitnames']
        elif dtype == 'bool':
            bits = [var]
        m = 2**len(bits)
        primed = stx.prime(var)
        pbits = [stx.prime(bit) for bit in bits]
        k = j + 1
        # copy attr
        dvars[var] = dict(d)
        dvars[primed] = dict(d)
        # update attr
        dvars[var].update(level=j, len=m, bitnames=bits)
        dvars[primed].update(level=k, len=m, bitnames=pbits)
    return dvars


def assert_primed_adjacent(prime, bdd):
    """Raise `AssertionError` if not neighbors.

    @param prime: map each unprimed to a primed var.
    @type prime: `dict(str: str)`
    @type bdd: `BDD`
    """
    # check adjacency of unprimed-primed pairs
    for x, y in prime.iteritems():
        i = bdd.level_of_var(x)
        j = bdd.level_of_var(y)
        assert abs(i - j) == 1, (
            'Variables "{x}" (level {i}) and '
            '"{y}" (level {j}) are not adjacent.\n'
            'Primed and unprimed vars must be adjacent in '
            'the given BDD.').format(x=x, y=y, i=i, j=j)


def _join(c):
    """Return `str` with one element of `c` per line."""
    return '\n'.join(str(x) for x in c)



"""Conversion of semi-symbolic to symbolic automata.

Re-implementation of `tulip.synth`.
"""
from __future__ import absolute_import
import logging
import warnings

from omega.logic import syntax as stx
from omega.symbolic.symbolic import Automaton


logger = logging.getLogger(__name__)


def graph_to_logic(g, nodevar, ignore_initial,
                   receptive=False, self_loops=False):
    """Flatten labeled graph to temporal logic formulae.

    @param g: `TransitionSystem`

    The attribute `g.owner` defines who selects the next node.
    The attribute `g.env_vars` determines who controls each variable.

    @param g: `TransitionSystem`
    @param nodevar: variable that stores current node
    @type nodevar: `str`
    @param ignore_initial: if `False`, then add initial condition
        using `g.initia_nodes`.
    @type ignore_initial: `bool`
    @param receptive: if `True`, then add assumptions to
        ensure receptiveness at each node.
    @param self_loops: if `True`, then add all self-loops

    @return: temporal formulae representing `g`.
    @rtype: `Automaton`
    """
    assert g.assert_consistent()
    assert len(g) > 0
    t = _vars_to_symbol_table(g, nodevar)
    # add primed copies -- same `dict` value
    dvars = dict(t)
    p, _ = _prime_dict(dvars)
    dvars.update(p)
    # convert to logic
    init = _init_from_ts(g.initial_nodes, nodevar,
                         dvars, ignore_initial)
    tmp_init, nodepred = _node_var_trans(g, nodevar, dvars)
    sys_tran = list()
    env_tran = list()
    if g.owner == 'sys':
        sys_init = init + tmp_init
        r = _sys_trans(g, nodevar, dvars)
        if self_loops:
            r = '({r}) | ((X {var}) = {var})'.format(
                r=r, var=nodevar)
        sys_tran.append(r)
        sys_tran.extend(nodepred)
        env_init = list()
        if receptive:
            r = _env_trans_from_sys_ts(g, nodevar, dvars)
            env_tran.append(r)
    elif g.owner == 'env':
        sys_init = list()
        env_init = init + tmp_init
        r = _env_trans(g, nodevar, dvars, self_loops)
        if self_loops:
            r = '({r}) | ((X {var}) = {var})'.format(
                r=r, var=nodevar)
        env_tran.append(r)
        env_tran.extend(nodepred)
    else:
        raise Exception('owner is neither "sys" nor "env".')
    a = Automaton()
    a.vars = t
    a.init['env'].extend(env_init)
    a.init['sys'].extend(sys_init)
    a.action['env'].extend(env_tran)
    a.action['sys'].extend(sys_tran)
    return a


def _vars_to_symbol_table(g, nodevar):
    """Return `dict` of integer and Boolean variables.

    Conforms to `openpromela.bitvector` input:
      - type: 'bool' | 'saturating'
      - dom: (min, max)
      - owner: 'env' | 'sys'
    """
    t = dict()
    for var, dom in g.vars.iteritems():
        # type and domain
        if dom == 'bool':
            dtype = 'bool'
            dom = None
        elif dom == 'boolean':
            raise Exception('replace "boolean" with "bool"')
        else:
            assert isinstance(dom, tuple), (var, dom)
            dtype = 'saturating'
        # owner
        if var in g.env_vars:
            owner = 'env'
        else:
            owner = 'sys'
        t[var] = dict(type=dtype, dom=dom, owner=owner)
    assert all(isinstance(u, int) for u in g)
    dtype = 'saturating'
    dom = (min(g), max(g))
    t[nodevar] = dict(type=dtype, dom=dom, owner=g.owner)
    return t


def _node_var_trans(g, nodevar, dvars):
    """Return data flow constraint on variables labeling nodes."""
    init = list()
    trans = list()
    # no AP labels ?
    if not dvars:
        return (init, trans)
    for u, d in g.nodes_iter(data=True):
        pre = _assign(nodevar, u, dvars)
        r = _to_action(d, dvars)
        if r == 'True':
            continue
        # initial node vars
        init.append('!({pre}) || ({r})'.format(pre=pre, r=r))
        # transitions of node vars
        trans.append('(X (({pre}) -> ({r})))'.format(pre=pre, r=r))
    return (init, trans)


def _init_from_ts(initial_nodes, nodevar,
                  dvars, ignore_initial=False):
    """Return initial condition."""
    if ignore_initial:
        return list()
    if not initial_nodes:
        raise Exception('Transition system without initial states.')
    return [stx.disj(_assign(nodevar, u, dvars) for u in initial_nodes)]


def _sys_trans(g, nodevar, dvars):
    """Convert transition relation to safety formula."""
    logger.debug('modeling sys transitions in logic')
    sys_trans = list()
    for u in g:
        pre = _assign(nodevar, u, dvars)
        # no successors ?
        if not g.succ.get(u):
            logger.debug('node: {u} is deadend !'.format(u=u))
            sys_trans.append('({pre}) -> (X False)'.format(pre=pre))
            continue
        post = list()
        for u, v, d in g.edges_iter(u, data=True):
            t = dict(d)
            t[stx.prime(nodevar)] = v
            r = _to_action(t, dvars)
            post.append(r)
        c = '({pre}) -> ({post})'.format(pre=pre, post=stx.disj(post))
        sys_trans.append(c)
    s = stx.conj(sys_trans, sep='\n')
    return s


def _env_trans_from_sys_ts(g, nodevar, dvars):
    """Return safety assumption to prevent env from blocking sys."""
    denv = {k: v for k, v in dvars.iteritems() if k in g.env_vars}
    env_trans = list()
    for u in g:
        # no successor states ?
        if not g.succ.get(u):
            continue
        # collect possible next env actions
        c = set()
        for u, w, d in g.edges_iter(u, data=True):
            t = _to_action(d, denv)
            if not t:
                continue
            c.add(t)
        # no next env actions ?
        if not c:
            continue
        post = stx.disj(c)
        pre = _assign(nodevar, u, dvars)
        env_trans.append('(({pre}) -> ({post}))'.format(
            pre=pre, post=post))
    s = stx.conj(env_trans, sep='\n')
    return s


def _env_trans(g, nodevar, dvars, self_loops):
    """Convert environment transitions to safety formula.

    @type g: `networkx.MultiDigraph`
    @param nodevar: name of variable representing current node
    @type nodevar: `str`
    @type dvars: `dict`
    """
    env_trans = list()
    for u in g:
        pre = _assign(nodevar, u, dvars)
        # no successors ?
        if not g.succ.get(u):
            env_trans.append('{pre} -> X(False)'.format(pre=pre))
            if not self_loops:
                warnings.warn(
                    'Environment dead-end found.\n'
                    'If sys can force env to dead-end,\n'
                    'then GR(1) assumption becomes False,\n'
                    'and spec trivially True.')
            continue
        post = list()
        sys = list()
        for u, v, d in g.out_edges_iter(u, data=True):
            # action
            t = dict(d)
            t[stx.prime(nodevar)] = v
            r = _to_action(t, dvars)
            post.append(r)
            # what sys vars ?
            t = {k: v for k, v in d.iteritems()
                 if k not in g.env_vars}
            r = _to_action(t, dvars)
            sys.append(r)
        # avoid sys winning env by blocking all edges
        # post.append(stx.conj_neg(sys))
        env_trans.append('({pre}) -> ({post})'.format(
            pre=pre, post=stx.disj(post)))
    s = stx.conj(env_trans, sep='\n')
    return s


def _to_action(d, dvars):
    """Return `str` conjoining assignments and `"formula"` in `d`.

    @param d: (partial) mapping from variables in `dvars`
        to values in their range, defined by `dvars`
    @type d: `dict`
    @type dvars: `dict`
    """
    c = list()
    if 'formula' in d:
        c.append(d['formula'])
    for k, v in d.iteritems():
        if k not in dvars:
            continue
        s = _assign(k, v, dvars)
        c.append(s)
    return stx.conj(c)


def _assign(k, v, dvars):
    """Return `str` of equality of variable `k` to value `v`.

    @type k: `str`
    @type v: `str` or `int`
    @type dvars: `dict`
    """
    dom = dvars[k]['dom']
    if isinstance(dom, tuple):
        s = '{k} = {v}'.format(k=k, v=v)
    elif isinstance(dom, (set, list)):
        s = '{k} = "{v}"'.format(k=k, v=v)
    elif dom == 'bool':
        s = '{k} <-> {v}'.format(k=k, v=v)
    else:
        raise Exception('domain is: {dom}'.format(dom=dom))
    return _pstr(s)


def _prime_dict(d):
    """Return `dict` with primed keys and `dict` mapping to them."""
    p = dict((stx.prime(k), d[k]) for k in d)
    d2p = {k: stx.prime(k) for k in d}
    return p, d2p


def _pstr(x):
    return '({x})'.format(x=x)



"""Fixpoint algorithms using decision diagrams.

- attractor computation (least fixpoint)
- control invariance (greatest fixpoint)
- controllable preimage
"""
import logging
from dd import bdd as _bdd


SYS = 'sys'  # default player for existential image
logger = logging.getLogger(__name__)


def attractor(env_action, sys_action, target, aut,
              evars=None, inside=None):
    """Return attractor for `target`.

    Keyword args as `ue_preimage`.

    @param inside: remain in this set
    """
    logger.info('++ attractor')
    # no free primed vars must remain after quantification
    # support = a.bdd.support(a.action)
    # support = {a.bdd.vars[k] for k in support}
    # primed_support = set(k for k in map(a.prime.get, support)
    # if k is not None)
    # unquantified = primed_support.difference(
    # a.uvars).difference(a.evars)
    # assert not unquantified, unquantified
    # ancestors
    bdd = aut.bdd
    q = target
    qold = None
    while q != qold:
        qold = q
        pred = ue_preimage(
            env_action, sys_action, q, aut, evars=evars)
        q = bdd.apply('or', qold, pred)
        if inside is not None:
            q = aut.bdd.apply('and', q, inside)
    logger.info('-- attractor')
    return q


def trap(env_action, sys_action, safe, aut,
         evars=None, unless=None):
    """Return subset of `safe` with contolled exit.

    @param unless: if `None`, then returned controlled invariant
        subset of `safe`. Otherwise, this defines an allowed set.
    @rtype: BDD node
    """
    logger.info('++ cinv')
    bdd = aut.bdd
    q = bdd.true  # if unless is not None, then q = safe is wrong
    qold = None
    while q != qold:
        qold = q
        pre = ue_preimage(env_action, sys_action, q, aut, evars=evars)
        q = bdd.apply('and', safe, pre)
        if unless is not None:
            q = bdd.apply('or', q, unless)
    logger.info('-- cinv')
    return q


def ue_preimage(env_action, sys_action, target, aut,
                evars=None):
    r"""Return controllable predecessor set.

    Preimage with alternating quantification.
    Quantifier order: If `aut.moore`:

      - \E \A, else
      - \A \E

    Implication causality: If `aut.plus_one`:

      - /\ sys_action
        /\ env_action => target

      - env_action => /\ sys_action
                      /\ target
    """
    # TODO: controllable predecessor operator implemented
    # efficiently like relational product
    bdd = aut.bdd
    if evars is None:
        evars = aut.epvars
        uvars = aut.upvars
    else:
        uvars = set(aut.unprime).difference(evars)
    u = bdd.rename(target, aut.prime)
    if aut.plus_one:
        u = bdd.apply('and', sys_action, u)
        u = bdd.apply('->', env_action, u)
    else:
        u = bdd.apply('->', env_action, u)
        u = bdd.apply('and', sys_action, u)
    if aut.moore:
        # \E \A
        u = bdd.forall(uvars, u)
        u = bdd.exist(evars, u)
    else:
        # \A \E
        u = bdd.exist(evars, u)
        u = bdd.forall(uvars, u)
    return u


def preimage(trans, target, qvars, automaton, forall):
    """Preimage with non-mixed quantification."""
    return _bdd.preimage(
        trans, target, automaton.prime,
        qvars, automaton.bdd, forall)


def descendants(source, constrain, a, future=True):
    """Existential descendants of `source` in `constrain`."""
    if future:
        q = ee_image(source, a)
    else:
        q = source
    qold = None
    while q != qold:
        post = ee_image(q, a)
        qold = q
        q = a.bdd.apply('or', qold, post)
        q = a.bdd.apply('and', q, constrain)
    return q


def ee_image(source, a):
    """Existential image."""
    u = a.action[SYS][0]
    v = source
    assert u in a.bdd
    qvars = a.uevars
    bdd = a.bdd
    u = bdd.apply('and', u, v)
    u = bdd.quantify(u, qvars, forall=False)
    u = bdd.rename(u, a.unprime)
    image = u
    return image



from omega.automata import TransitionSystem
from omega.symbolic import fixpoint as fx
from omega.symbolic import logicizer, symbolic


def test_descendants():
    g = TransitionSystem()
    g.add_path([0, 1, 2])
    a = logicizer.graph_to_logic(g, 'pc', True)
    symbolic.fill_blanks(a)
    aut = a.build()
    source = aut.add_expr('pc = 0')
    constrain = aut.bdd.true
    v = fx.descendants(source, constrain, aut)
    assert v == aut.bdd.add_expr('pc_0 ^ pc_1'), aut.bdd.to_expr(v)


def test_ee_image_only_sys():
    g = TransitionSystem()
    g.add_path([0, 1, 2])
    a = logicizer.graph_to_logic(g, 'pc', True)
    symbolic.fill_blanks(a)
    aut = a.build()
    u = aut.add_expr('pc = 0')
    v = fx.ee_image(u, aut)
    v_ = aut.add_expr('pc = 1')
    assert v == v_, a.bdd.to_expr(v)
    v = fx.ee_image(v, aut)
    v_ = aut.add_expr('pc = 2')
    assert v == v_, aut.bdd.to_expr(v)


def test_ue_image_no_constrain():
    g = TransitionSystem()
    g.vars = dict(x='bool')
    g.env_vars = {'x'}
    g.add_edge(0, 1, formula='x')
    g.add_edge(0, 2, formula='!x')
    a = logicizer.graph_to_logic(g, 'pc', True)
    symbolic.fill_blanks(a)
    aut = a.build()
    # source constrained to x
    source = aut.add_expr('x & (pc = 0)')
    u = fx.ee_image(source, aut)
    assert u == aut.add_expr('pc = 1')
    # source contains both x and ! x
    source = aut.add_expr('pc = 0')
    u = fx.ee_image(source, aut)
    assert u == aut.add_expr('(pc = 1) | (pc = 2)')


def test_ee_image():
    g = TransitionSystem()
    g.vars = dict(x='bool')
    g.env_vars = {'x'}
    g.add_edge(0, 1, formula='x')
    g.add_edge(0, 1, formula='!x')
    g.add_edge(0, 2, formula='x')
    a = logicizer.graph_to_logic(g, 'pc', True)
    symbolic.fill_blanks(a)
    aut = a.build()
    source = aut.add_expr('pc = 0')
    u = fx.ee_image(source, aut)
    u_ = aut.add_expr('(pc = 1) | (pc = 2)')
    assert u == u_, aut.bdd.to_expr(u)


def test_ue_preimage():
    a = symbolic.Automaton()
    a.vars = dict(x=dict(type='bool', owner='env'),
                  y=dict(type='bool', owner='sys'))
    a.action['sys'] = ["x' -> !y'"]
    symbolic.fill_blanks(a)
    aut = a.build()
    target = aut.add_expr('y')
    (env_action,) = aut.action['env']
    (sys_action,) = aut.action['sys']
    u = fx.ue_preimage(env_action, sys_action, target, aut)
    assert u == aut.bdd.false, aut.bdd.to_expr(u)
    u = fx.ue_preimage(env_action, sys_action, target, aut,
                       evars=aut.uepvars)
    assert u == aut.bdd.true, aut.bdd.to_expr(u)


if __name__ == '__main__':
    test_ue_preimage()
